import numpy as np
import binning_utils


def power_law_spectrum(energy, flux_density, spectral_index, pivot_energy):
    """
    power-law:
    f(energy) = flux_density * (energy/pivot_energy) ** spectral_index
    """
    return flux_density * (energy / pivot_energy) ** (spectral_index)


def estimate_signal_rate_for_power_law_per_s(
    signal_area_m2,
    energy_bin_edges_GeV,
    power_law_flux_density_per_m2_per_GeV_per_s,
    power_law_spectral_index,
    power_law_pivot_energy_GeV,
):
    """
    Estimate the rate of signal-counts in the on-region $R_S / s^{-1}$
    given the emission's energy-spectrum is a power-law:

    f(energy) = flux_density * (energy/pivot_energy) ** spectral_index

    Parameters
    ----------
    signal_area_m2 : array of N floats / m^2
        The signal's effective area in the on-region.
    energy_bin_edges_GeV : array of (N+1) floats / GeV
        The edges of the energy-bins used for the signal's effective area.
    power_law_flux_density_per_m2_per_GeV_per_s : float / m^{-2} s^{-1} (GeV)^{-1}
        The power-law's flux-density.
    power_law_spectral_index : float / 1
        The power-law's spectral-index.
    power_law_pivot_energy_GeV : float / GeV
        The power-law's pivot-energy.

    Returns
    -------
    The signal-rate $R_S$ : float
    """
    assert np.all(signal_area_m2 >= 0.0)
    assert len(energy_bin_edges_GeV) == len(signal_area_m2) + 1
    assert np.all(energy_bin_edges_GeV > 0.0)
    assert np.all(np.gradient(energy_bin_edges_GeV) > 0.0)

    assert power_law_flux_density_per_m2_per_GeV_per_s > 0.0
    assert power_law_pivot_energy_GeV > 0.0

    energy_bin_centers_GeV = binning_utils.centers(
        bin_edges=energy_bin_edges_GeV,
    )
    energy_bin_width_GeV = binning_utils.widths(
        bin_edges=energy_bin_edges_GeV,
    )

    differential_flux_per_m2_per_s_per_GeV = power_law_spectrum(
        energy=energy_bin_centers_GeV,
        flux_density=power_law_flux_density_per_m2_per_GeV_per_s,
        spectral_index=power_law_spectral_index,
        pivot_energy=power_law_pivot_energy_GeV,
    )

    differential_signal_rate_per_s_per_GeV = (
        differential_flux_per_m2_per_s_per_GeV * signal_area_m2
    )

    signal_rate_per_s = np.sum(
        differential_signal_rate_per_s_per_GeV * energy_bin_width_GeV
    )
    return signal_rate_per_s


def _relative_ratio(a, b):
    return np.abs(a - b) / (0.5 * (a + b))


def estimate_flux_densities_of_critical_power_laws(
    signal_area_m2,
    energy_bin_edges_GeV,
    critical_signal_rate_per_s,
    power_law_spectral_indices,
    power_law_pivot_energy_GeV=1.0,
    margin=1e-2,
    upper_flux_density_per_m2_per_GeV_per_s=1e6,
    max_num_iterations=10000,
):
    """
    Estimates the flux-density of the critical power-laws which are still
    detectable.

    power-law:
    f(energy) = flux_density * (energy/pivot_energy) ** spectral_index

    Parameters
    ----------
    signal_area_m2 : array of N floats
        The signal's effective area in the on-region.
    energy_bin_edges_GeV : list of (N+1) floats
        The edges of the energy-bins used for the effective area.
    critical_signal_rate_per_s : float
        The critical rate of signal in the on-region which is required
        to claim a detection.
    power_law_spectral_indices : list of floats
        A list of the power-law's spectral-indices.
    power_law_pivot_energy_GeV : float
        Same for all power-laws.
    margin : float
        Stopping-criteria for iterative search of flux-density. When the
        relative ratio of the critical_signal_rate_per_s and the acual
        signal_rate_per_s caused by a particular power-law is below this
        margin, the search is complete.
    upper_flux_density_per_m2_per_GeV_per_s : float
        Starting point for iterative search of flux-density. This should be
        larger than the expected flux.
    max_num_iterations : int
        Stop the iterative search of flux-density in any case after
        this many iterations.

    Returns
    -------
    power_law_flux_densities : list of floats
    """
    assert critical_signal_rate_per_s > 0.0

    assert power_law_pivot_energy_GeV > 0.0
    assert margin > 0.0
    assert upper_flux_density_per_m2_per_GeV_per_s > 0.0
    assert max_num_iterations > 0

    power_law_flux_densities = []

    for i, power_law_spectral_index in enumerate(power_law_spectral_indices):
        flux_dens = float(upper_flux_density_per_m2_per_GeV_per_s)

        iteration = 0
        while True:
            assert iteration < max_num_iterations

            signal_rate_per_s = estimate_signal_rate_for_power_law_per_s(
                signal_area_m2=signal_area_m2,
                energy_bin_edges_GeV=energy_bin_edges_GeV,
                power_law_flux_density_per_m2_per_GeV_per_s=flux_dens,
                power_law_spectral_index=power_law_spectral_index,
                power_law_pivot_energy_GeV=power_law_pivot_energy_GeV,
            )

            ratio = _relative_ratio(
                signal_rate_per_s, critical_signal_rate_per_s
            )

            if ratio < margin:
                break

            rr = ratio / 3
            if signal_rate_per_s > critical_signal_rate_per_s:
                flux_dens *= 1 - rr
            else:
                flux_dens *= 1 + rr

            iteration += 1

        power_law_flux_densities.append(float(flux_dens))
    return power_law_flux_densities


def _find_intersection_two_lines(b1, m1, b2, m2):
    """
    Find the intersection of two affine functions:
    f1(x) = m1 * x + b1, and f2(x) = m2 * x + b2 in x.

    Parameters
    ----------
    m1 : float
        Slope of f1(x).
    b1 : float
        Support of f1(x).
    m2 : float
        Slope of f2(x).
    b2 : float
        Support of f2(x).

    Returns
    -------
    x : float
        Intersection of f1(x) and f2(x).
    """
    return (b2 - b1) / (m1 - m2)


def _estimate_tangent_of_consecutive_power_laws(A_ns, G_ns):
    """
    Estimate the curve described by the intersection-points of two
    consecutive power-laws in a list of N power-laws [f_0(x), ..., f_N(x)].

    f_0(x) = A_0 * x ** G_0,
           .
           .
    f_N(x) = A_N * x ** G_N

    Parameters
    ----------
    A_ns : list of floats
        N power-laws normalizations: A_ns = [A_0, A_1, A_2, ... , A_N]
    G_ns : list of floats
        N power-laws exponents: G_ns = [G_0, G_1, G_2, ... , G_N]

    Returns
    -------
    (x, y) : (array of floats, array of floats)
        List of N-1 intersections for N power-laws.
    """
    assert len(A_ns) == len(G_ns)
    num = len(A_ns)
    assert num >= 2

    log10_A_ns = np.log10(np.array(A_ns))
    G_ns = np.array(G_ns)

    x = []
    y = []
    for i in range(num - 1):
        log10_x = _find_intersection_two_lines(
            b1=log10_A_ns[i],
            m1=G_ns[i],
            b2=log10_A_ns[i + 1],
            m2=G_ns[i + 1],
        )
        log10_y = log10_A_ns[i] + G_ns[i] * (log10_x)
        x.append(10**log10_x)
        y.append(10**log10_y)
    return (np.array(x), np.array(y))


def estimate_tangent_of_consecutive_power_laws(
    flux_densities, spectral_indices
):
    return _estimate_tangent_of_consecutive_power_laws(
        A_ns=flux_densities,
        G_ns=spectral_indices,
    )


def estimate_spectral_exclusion_zone(
    signal_area_m2,
    energy_bin_edges_GeV,
    critical_signal_rate_per_s,
    power_law_spectral_indices=np.linspace(start=-5, stop=-0.5, num=137),
    power_law_pivot_energy_GeV=1.0,
):
    """
    Estimates a curve in the space of differential flux and energy.
    All power-laws below this curve can not be detected.

    Parameters
    ----------
    signal_area_m2 : list of N floats
        The effective area where signal is collected in the on-region.
    energy_bin_edges_GeV : list of (N+1) floats
        The edges of the energy-bins used for the effective area.
    critical_signal_rate_per_s : float
        The minimal rate of signal in the on-region $R_S$ required to
        claim a detection.
    power_law_spectral_indices : list of floats
    """
    assert critical_signal_rate_per_s > 0.0

    power_law_spectral_indices = np.array(power_law_spectral_indices)
    assert len(power_law_spectral_indices) >= 2
    assert np.all(np.gradient(power_law_spectral_indices) > 0.0)

    assert power_law_pivot_energy_GeV > 0.0

    power_law_flux_densities = estimate_flux_densities_of_critical_power_laws(
        signal_area_m2=signal_area_m2,
        energy_bin_edges_GeV=energy_bin_edges_GeV,
        critical_signal_rate_per_s=critical_signal_rate_per_s,
        power_law_spectral_indices=power_law_spectral_indices,
        power_law_pivot_energy_GeV=power_law_pivot_energy_GeV,
    )

    (
        energy_GeV,
        diff_flux_per_m2_per_GeV_per_s,
    ) = estimate_tangent_of_consecutive_power_laws(
        flux_densities=power_law_flux_densities,
        spectral_indices=power_law_spectral_indices,
    )
    return energy_GeV, diff_flux_per_m2_per_GeV_per_s



import lima1983analysis
import propagate_uncertainties as pru


def estimate_critical_signal_rate(
    background_rate_onregion_in_scenario_per_s,
    background_rate_onregion_in_scenario_per_s_au,
    onregion_over_offregion_ratio,
    observation_time_s,
    instrument_systematic_uncertainty_relative,
    detection_threshold_std,
    estimator_statistics="LiMaEq17",
    combiner_statistics_systematics="hypot",
):
    """
    "Algorithm C"
    Estimate the critical rate of signal counts $R_S$ in the on-rerion
    required to claim a detection.

    Parameters
    ----------
    background_rate_onregion_in_scenario_per_s : float / s^{-1}
        Expected rate of background counts in the on-region for a certain
        scenario.
        This is $\\hat{R}_B = \\hat{N}_B / T_\\text{obs}$.
    background_rate_onregion_in_scenario_per_s_au : float / s^{-1}
        Absulute uncertainty.
    onregion_over_offregion_ratio : float / 1
        Ratio of on- over off-region.
        This is $\\alpha$
    observation_time_s : float / s
        Effective time of observation, $T_\\text{obs}$.
    instrument_systematic_uncertainty_relative : float / 1
        The instrument's relative systematic uncertaintiy.
    detection_threshold_std : float / 1
        Significance in stdandard deviations.
    estimator_statistics : str
        Statistical estimator for required num. signal-counts in onregion.
        ["sqrt", "LiMaEq9", "LiMaEq17"]
    combiner_statistics_systematics : str
        ["max", "hypot"]

    Returns
    -------
    R_S, R_S_au : (float, float)
        The signal's minimal rate in the on-region required to claim a
        detection.
    """
    hatR_B = float(background_rate_onregion_in_scenario_per_s)
    hatR_B_au = float(background_rate_onregion_in_scenario_per_s_au)
    alpha = float(onregion_over_offregion_ratio)
    T_obs = float(observation_time_s)
    S = float(detection_threshold_std)
    U_sys_rel_unc = float(instrument_systematic_uncertainty_relative)

    assert hatR_B >= 0.0
    assert hatR_B_au >= 0.0
    assert T_obs > 0.0
    assert alpha > 0.0
    assert U_sys_rel_unc >= 0.0
    assert S > 0.0

    hatN_B, hatN_B_au = pru.multiply(x=hatR_B, x_au=hatR_B_au, y=T_obs, y_au=0)
    N_off, N_off_au = pru.divide(x=hatN_B, x_au=hatN_B_au, y=alpha, y_au=0)

    if estimator_statistics == "sqrt":
        N_S_stat, N_S_stat_au = estimator_sqrt(
            N_off=N_off,
            N_off_au=N_off_au,
            alpha=alpha,
            S=S,
        )
    elif estimator_statistics == "LiMaEq9":
        N_S_stat, N_S_stat_au = estimator_lima1983analysis_N_s_eq9(
            N_off=N_off,
            N_off_au=N_off_au,
            alpha=alpha,
            S=S,
        )
    elif estimator_statistics == "LiMaEq17":
        N_S_stat, N_S_stat_au = estimator_lima1983analysis_N_s_eq17(
            N_off=N_off,
            N_off_au=N_off_au,
            alpha=alpha,
            S=S,
            margin=1e-4,
            max_num_iterations=10 * 1000,
        )
    else:
        raise KeyError(
            "Unknown estimator for statistics: '{:s}'".format(
                estimator_statistics
            )
        )

    # estimate the required signal-rate to satisfy the systematic uncertainty
    N_s_sys, N_s_sys_au = pru.multiply(
        x=(S * U_sys_rel_unc), x_au=0.0, y=hatN_B, y_au=hatN_B_au
    )

    # combine statistics and systematics
    if combiner_statistics_systematics == "max":
        N_s, N_s_au = pru.max(
            x=[N_S_stat, N_s_sys], x_au=[N_S_stat_au, N_s_sys_au]
        )
    elif combiner_statistics_systematics == "hypot":
        N_s, N_s_au = pru.hypot(
            x=N_S_stat, x_au=N_S_stat_au, y=N_s_sys, y_au=N_s_sys_au
        )
    else:
        raise KeyError(
            "Unknown mixing of statistics and systematics: '{:s}'".format(
                combiner_statistics_systematics
            )
        )

    R_s, R_s_au = pru.divide(x=N_s, x_au=N_s_au, y=T_obs, y_au=0.0)
    return R_s, R_s_au


def estimator_sqrt(N_off, N_off_au, alpha, S):
    N_off_std, N_off_std_au = pru.sqrt(x=N_off, x_au=N_off_au)
    N_on_std, N_on_std_au = pru.multiply(
        x=N_off_std, x_au=N_off_std_au, y=alpha, y_au=0.0
    )
    return pru.multiply(x=S, x_au=0.0, y=N_on_std, y_au=N_on_std_au)


def estimator_lima1983analysis_N_s_eq9(
    N_off, N_off_au, alpha, S, numeric_derivative_epsilon_relative=1e-2
):
    lima_eq9 = lima1983analysis.estimate_N_s_eq9
    N_off_plus = N_off * (1.0 + numeric_derivative_epsilon_relative)
    N_off_minus = N_off * (1.0 - numeric_derivative_epsilon_relative)
    N_s_plus = lima_eq9(N_off=N_off_plus, alpha=alpha, S=S)
    N_s_minus = lima_eq9(N_off=N_off_minus, alpha=alpha, S=S)
    dN_s_dN_off = derivative(
        f_plus=N_s_plus,
        f_minus=N_s_minus,
        x_plus=N_off_plus,
        x_minus=N_off_minus,
    )
    N_s = lima_eq9(N_off=N_off, alpha=alpha, S=S)
    N_s_au = pru.auN(dfdx=[dN_s_dN_off], x_au=[N_off_au])
    return N_s, N_s_au


def estimator_lima1983analysis_N_s_eq17(
    N_off,
    N_off_au,
    alpha,
    S,
    margin=1e-4,
    max_num_iterations=10 * 1000,
    numeric_derivative_epsilon_relative=1e-2,
):
    lima_eq17 = lima1983analysis.estimate_N_s_eq17
    N_off_plus = N_off * (1.0 + numeric_derivative_epsilon_relative)
    N_off_minus = N_off * (1.0 - numeric_derivative_epsilon_relative)
    N_s_plus = lima_eq17(
        N_off=N_off_plus,
        alpha=alpha,
        S=S,
        margin=margin,
        max_num_iterations=max_num_iterations,
    )
    N_s_minus = lima_eq17(
        N_off=N_off_minus,
        alpha=alpha,
        S=S,
        margin=margin,
        max_num_iterations=max_num_iterations,
    )
    dN_s_dN_off = derivative(
        f_plus=N_s_plus,
        f_minus=N_s_minus,
        x_plus=N_off_plus,
        x_minus=N_off_minus,
    )
    N_s = lima_eq17(
        N_off=N_off,
        alpha=alpha,
        S=S,
        margin=margin,
        max_num_iterations=max_num_iterations,
    )
    N_s_au = pru.auN(dfdx=[dN_s_dN_off], x_au=[N_off_au])
    return N_s, N_s_au


def derivative(f_plus, f_minus, x_plus, x_minus):
    df = f_plus - f_minus
    dx = x_plus - x_minus
    return df / dx



import numpy as np
import propagate_uncertainties as pru
import binning_utils
from .. import critical_rate
from . import scenarios


def estimate_differential_sensitivity(
    energy_bin_edges_GeV,
    signal_area_in_scenario_m2,
    signal_area_in_scenario_m2_au,
    critical_signal_rate_in_scenario_per_s,
    critical_signal_rate_in_scenario_per_s_au,
):
    """
    Estimates the differential flux-sensitivity in N energy-bin for a
    certain scenario.

    Parameters
    ----------
    energy_bin_edges_GeV : array of (N+1) floats / GeV
        Edges of the energy-bins.
    signal_area_in_scenario_m2 : array of N floats / m^{2}
        Signal's effective area for each energy-bin w.r.t. to the applied
        scenario.
    signal_area_in_scenario_m2_au : array of N floats / m^{2}
        Absolute uncertainty.
    critical_signal_rate_in_scenario_per_s : array of N floats / s^{-1}
        Signal's minimal rate to claim a detection for each energy-bin w.r.t.
        the scenario, the same scenario as applied in
        signal_area_in_scenario_m2.
    critical_signal_rate_in_scenario_per_s_au : array of N floats / s^{-1}
        Absolute uncertainty.
    Returns
    -------
    differential_flux_vs_energy : array of N floats / m^{-2} s^{-1} (GeV)^{-1}
        The minimal differential flux required to claim a detection in a
        certain scenario.
    """
    num_energy_bins = len(energy_bin_edges_GeV) - 1
    assert num_energy_bins >= 1, "Need at least two bin-edges."

    assert np.all(energy_bin_edges_GeV > 0.0), "Energy must be positive."
    assert np.all(
        np.gradient(energy_bin_edges_GeV) > 0.0
    ), "Energy bin-edges must be increasing."

    assert num_energy_bins == len(signal_area_in_scenario_m2)
    assert num_energy_bins == len(critical_signal_rate_in_scenario_per_s)

    # work
    # ----
    dE_GeV = binning_utils.widths(bin_edges=energy_bin_edges_GeV)
    dVdE_per_s_per_m2_per_GeV = np.nan * np.ones(num_energy_bins)
    dVdE_per_s_per_m2_per_GeV_au = np.nan * np.ones(num_energy_bins)

    for ebin in range(num_energy_bins):
        if signal_area_in_scenario_m2[ebin] > 0:
            dV_per_s_per_m2, dV_per_s_per_m2_au = pru.divide(
                x=critical_signal_rate_in_scenario_per_s[ebin],
                x_au=critical_signal_rate_in_scenario_per_s_au[ebin],
                y=signal_area_in_scenario_m2[ebin],
                y_au=signal_area_in_scenario_m2_au[ebin],
            )
        else:
            dV_per_s_per_m2 = np.nan
            dV_per_s_per_m2_au = np.nan

        (
            dVdE_per_s_per_m2_per_GeV[ebin],
            dVdE_per_s_per_m2_per_GeV_au[ebin],
        ) = pru.divide(
            x=dV_per_s_per_m2,
            x_au=dV_per_s_per_m2_au,
            y=dE_GeV[ebin],
            y_au=0.0,
        )

    return dVdE_per_s_per_m2_per_GeV, dVdE_per_s_per_m2_per_GeV_au


SCENARIOS = {
    "blue": {"energy_axes_label": "", "obsolete_key": "perfect_energy"},
    "yellow": {
        "energy_axes_label": "reco. ",
        "obsolete_key": "broad_spectrum",
    },
    "green": {"energy_axes_label": "", "obsolete_key": "line_spectrum"},
    "black": {"energy_axes_label": "", "obsolete_key": "bell_spectrum"},
}


def init_scenario_matrices_for_signal_and_background(
    probability_reco_given_true,
    probability_reco_given_true_au,
    scenario_key="yellow",
):
    """
    Make the scenario dependent matrices G for signal and B for background
    based on the energy-confusion of the signal.

    Parameters
    ----------
    probability_reco_given_true : array(N, N) / 1
        Energy-confusion-matrix.
        Conditional probability to measure E_reco given the E_true.
        ax0: true
        ax1: reco
        For the signal, e.g. gamma-rays.
    probability_reco_given_true_au : array(N, N) / 1
        Absolute uncertainty.
    scenario_key : str
        Key of the chosen scenario.
    """
    shape = probability_reco_given_true.shape
    assert probability_reco_given_true_au.shape == shape

    if scenario_key == "blue":
        G = np.eye(N=shape[0])
        G_au = np.zeros(shape=shape)  # zero uncertainty

        B = np.eye(N=shape[0])
        B_au = np.zeros(shape=shape)

    elif scenario_key == "yellow":
        G = np.array(probability_reco_given_true)
        G_au = np.array(probability_reco_given_true_au)  # adopt as is

        B = np.eye(N=shape[0])
        B_au = np.zeros(shape=shape)

    elif scenario_key == "green":
        # only the diagonal
        eye = np.eye(N=shape[0])
        G = eye * np.diag(probability_reco_given_true)
        G_au = eye * np.diag(probability_reco_given_true_au)

        B = np.eye(N=shape[0])
        B_au = np.zeros(shape=shape)

    elif scenario_key == "black":
        containment = 0.68
        G = containment * np.eye(N=shape[0])  # true energy for gammas
        G_au = np.zeros(shape=shape)  # zero uncertainty

        B = scenarios.black.init_matrix_B(
            probability_reco_given_true=probability_reco_given_true,
            containment=containment,
        )
        B_au = np.zeros(shape=shape)

    else:
        raise KeyError("Unknown scenario_key: '{:s}'".format(scenario_key))

    return {
        "G_matrix": G,
        "G_matrix_au": G_au,
        "B_matrix": B,
        "B_matrix_au": B_au,
        "energy_axes_label": SCENARIOS[scenario_key]["energy_axes_label"],
    }


def estimate_critical_signal_rate_vs_energy(
    background_rate_onregion_in_scenario_per_s,
    background_rate_onregion_in_scenario_per_s_au,
    onregion_over_offregion_ratio,
    observation_time_s,
    instrument_systematic_uncertainty_relative,
    detection_threshold_std,
    estimator_statistics,
):
    critical_signal_rate_per_s = np.nan * np.ones(
        shape=background_rate_onregion_in_scenario_per_s.shape
    )
    critical_signal_rate_per_s_au = np.nan * np.ones(
        shape=background_rate_onregion_in_scenario_per_s.shape
    )

    for ebin in range(len(background_rate_onregion_in_scenario_per_s)):
        if background_rate_onregion_in_scenario_per_s[ebin] > 0.0:
            (
                critical_signal_rate_per_s[ebin],
                critical_signal_rate_per_s_au[ebin],
            ) = critical_rate.estimate_critical_signal_rate(
                background_rate_onregion_in_scenario_per_s=background_rate_onregion_in_scenario_per_s[
                    ebin
                ],
                background_rate_onregion_in_scenario_per_s_au=background_rate_onregion_in_scenario_per_s_au[
                    ebin
                ],
                onregion_over_offregion_ratio=onregion_over_offregion_ratio,
                observation_time_s=observation_time_s,
                instrument_systematic_uncertainty_relative=instrument_systematic_uncertainty_relative,
                detection_threshold_std=detection_threshold_std,
                estimator_statistics=estimator_statistics,
            )
        else:
            critical_signal_rate_per_s[ebin] = float("nan")
            critical_signal_rate_per_s_au[ebin] = float("nan")
    return critical_signal_rate_per_s, critical_signal_rate_per_s_au


def apply_scenario_to_signal_effective_area(
    signal_area_m2,
    signal_area_m2_au,
    scenario_G_matrix,
    scenario_G_matrix_au,
):
    """
    Apply scenario (matrix G) to the signal's effective area.

    Parameters
    ----------
    signal_area_m2 : array of N floats / m^{2}
        Signal's effective area for each energy-bin in true energy.
    signal_area_m2_au : array of N floats / m^{2}
        Absolute uncertainty of signal_area_m2.
    scenario_G_matrix : array (N x N) / 1
        The scenario's matrix 'G' to confuse the signal's effective area.
    scenario_G_matrix_au : array (N x N) / 1
        Absolute uncertainty of scenario_G_matrix.

    Returns
    -------
    scenario_signal_area_m2, scenario_signal_area_m2_au
    """
    Atrue = signal_area_m2
    Atrue_au = signal_area_m2_au
    G = scenario_G_matrix
    G_au = scenario_G_matrix_au

    assert len(Atrue) == len(Atrue_au)
    assert np.all(Atrue >= 0)
    assert np.all(Atrue_au >= 0)
    N = len(Atrue)

    assert G.shape == G_au.shape
    assert G.shape[0] == G.shape[1]
    assert np.all(G >= 0)
    assert np.all(G_au >= 0)
    assert G.shape[0] == N

    Ascenario = np.zeros(N)
    Ascenario_au = np.zeros(N)

    for ereco in range(N):
        tmp = np.zeros(N)
        tmp_au = np.zeros(N)

        for etrue in range(N):
            tmp[etrue], tmp_au[etrue] = pru.prod(
                x=[
                    G[etrue, ereco],
                    Atrue[etrue],
                ],
                x_au=[
                    G_au[etrue, ereco],
                    Atrue_au[etrue],
                ],
            )

        Ascenario[ereco], Ascenario_au[ereco] = pru.sum(x=tmp, x_au=tmp_au)
    return Ascenario, Ascenario_au


def apply_scenario_to_background_rate(
    rate_in_reco_energy_per_s,
    rate_in_reco_energy_per_s_au,
    scenario_B_matrix,
    scenario_B_matrix_au,
):
    """
    Apply scenario (matrix B) to the background's rate in reco. energy.

    Parameters
    ----------
    rate_in_reco_energy_per_s : array of N floats / m^{2}
        Background's rate for each energy-bin in reco energy.
    rate_in_reco_energy_per_s_au : array of N floats / m^{2}
        Absolute uncertainty of rate_in_reco_energy_per_s.
    scenario_B_matrix : array (N x N) / 1
        The scenario's matrix 'B' to mask the energy-bins where
        background is collected in.
    scenario_B_matrix_au : array (N x N) / 1
        Absolute uncertainty of scenario_G_matrix.

    Returns
    -------
    rate_scenario_per_s, rate_scenario_per_s_au
        The scenario's rate of background in reco. energy.
    """
    Rreco = rate_in_reco_energy_per_s
    Rreco_au = rate_in_reco_energy_per_s_au
    B = scenario_B_matrix
    B_au = scenario_B_matrix_au

    assert len(Rreco) == len(Rreco_au)
    assert np.all(Rreco >= 0)
    assert np.all(Rreco_au >= 0)
    N = len(Rreco)

    assert B.shape == B_au.shape
    assert B.shape[0] == B.shape[1]
    assert np.all(B >= 0)
    assert np.all(B_au >= 0)
    assert B.shape[0] == N

    Rscenario = np.zeros(N)
    Rscenario_au = np.zeros(N)

    for ereco in range(N):
        tmp = np.zeros(N)
        tmp_au = np.zeros(N)

        for etrue in range(N):
            tmp[etrue], tmp_au[etrue] = pru.prod(
                x=[
                    B[ereco, etrue],
                    Rreco[etrue],
                ],
                x_au=[
                    B_au[ereco, etrue],
                    Rreco_au[etrue],
                ],
            )

        Rscenario[ereco], Rscenario_au[ereco] = pru.sum(x=tmp, x_au=tmp_au)
    return Rscenario, Rscenario_au


def assert_energy_reco_given_true_ax0true_ax1reco_is_normalized(
    energy_reco_given_true_ax0true_ax1reco, margin=1e-2
):
    M = energy_reco_given_true_ax0true_ax1reco
    assert M.shape[0] == M.shape[1]
    num_energy_bins = M.shape[0]

    for etrue in range(num_energy_bins):
        check = np.sum(M[etrue, :])
        if check > 0:
            assert (
                (1.0 - margin) < check < (1.0 + margin)
            ), "Expected sum(P(reco|true)) = 1, but it's {:f}".format(check)


def estimate_rate_in_reco_energy(
    energy_bin_edges_GeV,
    acceptance_m2_sr,
    acceptance_m2_sr_au,
    differential_flux_per_m2_per_sr_per_s_per_GeV,
    differential_flux_per_m2_per_sr_per_s_per_GeV_au,
    energy_reco_given_true_ax0true_ax1reco,
    energy_reco_given_true_ax0true_ax1reco_au,
):
    num_energy_bins = len(energy_bin_edges_GeV) - 1
    dE_GeV = binning_utils.widths(bin_edges=energy_bin_edges_GeV)
    dE_GeV_au = np.zeros(num_energy_bins)  # abs. uncertainty is zero

    rate_in_reco_energy_per_s = np.zeros(num_energy_bins)
    rate_in_reco_energy_per_s_au = np.zeros(num_energy_bins)

    for ereco in range(num_energy_bins):
        _summands = np.zeros(num_energy_bins)
        _summands_au = np.zeros(num_energy_bins)
        for etrue in range(num_energy_bins):
            _multiplicands = [
                differential_flux_per_m2_per_sr_per_s_per_GeV[etrue],
                energy_reco_given_true_ax0true_ax1reco[etrue, ereco],
                acceptance_m2_sr[etrue],
                dE_GeV[etrue],
            ]
            _multiplicands_au = [
                differential_flux_per_m2_per_sr_per_s_per_GeV_au[etrue],
                energy_reco_given_true_ax0true_ax1reco_au[etrue, ereco],
                acceptance_m2_sr_au[etrue],
                dE_GeV_au[etrue],
            ]
            (
                _summands[etrue],
                _summands_au[etrue],
            ) = pru.prod(x=_multiplicands, x_au=_multiplicands_au)
        (
            rate_in_reco_energy_per_s[ereco],
            rate_in_reco_energy_per_s_au[ereco],
        ) = pru.sum(x=_summands, x_au=_summands_au)

    return rate_in_reco_energy_per_s, rate_in_reco_energy_per_s_au


def estimate_rate_in_true_energy(
    energy_bin_edges_GeV,
    acceptance_m2_sr,
    acceptance_m2_sr_au,
    differential_flux_per_m2_per_sr_per_s_per_GeV,
    differential_flux_per_m2_per_sr_per_s_per_GeV_au,
):
    num_energy_bins = len(energy_bin_edges_GeV) - 1
    dE_GeV = binning_utils.widths(bin_edges=energy_bin_edges_GeV)
    dE_GeV_au = np.zeros(num_energy_bins)  # abs. uncertainty is zero

    rate_in_true_energy_per_s = np.zeros(num_energy_bins)
    rate_in_true_energy_per_s_au = np.zeros(num_energy_bins)

    for etrue in range(num_energy_bins):
        (
            rate_in_true_energy_per_s[etrue],
            rate_in_true_energy_per_s_au[etrue],
        ) = pru.prod(
            x=[
                differential_flux_per_m2_per_sr_per_s_per_GeV[etrue],
                acceptance_m2_sr[etrue],
                dE_GeV[etrue],
            ],
            x_au=[
                differential_flux_per_m2_per_sr_per_s_per_GeV_au[etrue],
                acceptance_m2_sr_au[etrue],
                dE_GeV_au[etrue],
            ],
        )

    return rate_in_true_energy_per_s, rate_in_true_energy_per_s_au


def assert_integral_rates_are_similar_in_reco_and_true_energy(
    rate_in_reco_energy_per_s,
    rate_in_true_energy_per_s,
    margin=0.3,
):
    """
    Integral rate over all energy-bins must not change (much) under
    energy migration.
    """
    integral_Rtrue = np.sum(rate_in_true_energy_per_s[:])
    integral_Rreco = np.sum(rate_in_reco_energy_per_s[:])
    assert (1.0 - margin) < integral_Rtrue / integral_Rreco < (1.0 + margin), (
        "Expected integral rates in both true and reco. energy to be "
        "roughly the same. But they are not."
    )



import astropy
from astropy.io import fits
import scipy
from scipy.interpolate import interp2d
import binning_utils as bu
import numpy as np


def average_field_of_view_radial(X, theta_bin_edges_deg, roi_opening_deg):
    avg_shape = X.shape[1:]
    X_avg = np.zeros(avg_shape)
    total_theta_solid_angle_deg2 = 0.0
    for t in range(len(theta_bin_edges_deg) - 1):
        theta_start_deg = theta_bin_edges_deg[t]
        theta_stop_deg = theta_bin_edges_deg[t + 1]
        theta_solid_angle_deg2 = np.pi * (
            theta_stop_deg**2 - theta_start_deg**2
        )
        theta_deg = 0.5 * (theta_start_deg + theta_stop_deg)
        if theta_deg <= roi_opening_deg:
            total_theta_solid_angle_deg2 += theta_solid_angle_deg2
            X_avg += theta_solid_angle_deg2 * X[t, :]

    X_avg /= total_theta_solid_angle_deg2
    return X_avg


def average_field_of_view_grid(
    X, detx_bin_edges_deg, dety_bin_edges_deg, roi_opening_deg
):
    avg_shape = X.shape[2:]
    X_avg = np.zeros(avg_shape)

    N_detx = len(detx_bin_edges_deg) - 1
    N_dety = len(dety_bin_edges_deg) - 1

    total_solid_angle = 0

    for x in range(N_detx):
        for y in range(N_dety):
            detx_start_deg = detx_bin_edges_deg[x]
            detx_stop_deg = detx_bin_edges_deg[x + 1]
            dety_start_deg = dety_bin_edges_deg[x]
            dety_stop_deg = dety_bin_edges_deg[x + 1]

            detx_deg = 0.5 * (detx_start_deg + detx_stop_deg)
            dety_deg = 0.5 * (dety_start_deg + dety_stop_deg)

            det_solid_angle = (detx_stop_deg - detx_start_deg) * (
                dety_stop_deg - dety_start_deg
            )
            if np.hypot(detx_deg, dety_deg) < roi_opening_deg:
                total_solid_angle += det_solid_angle
                X_avg += det_solid_angle * X[x, y, :]

    X_avg /= total_solid_angle
    return X_avg


def find_common_energy_bin_edges(components, num_bins_per_decade):
    multiple_edges = [
        components[k]["energy_bin_edges_GeV"] for k in components
    ]

    E_start = bu.max_lowest_edge(multiple_edges=multiple_edges)
    E_stop = bu.min_highest_edge(multiple_edges=multiple_edges)

    (start_decade, start_bin) = bu.power10.find_upper_decade_and_bin(
        x=E_start, num_bins_per_decade=num_bins_per_decade
    )
    (stop_decade, stop_bin) = bu.power10.find_lower_decade_and_bin(
        x=E_stop, num_bins_per_decade=num_bins_per_decade
    )

    return bu.power10.space(
        start_decade=start_decade,
        start_bin=start_bin,
        stop_decade=stop_decade,
        stop_bin=stop_bin,
        num_bins_per_decade=num_bins_per_decade,
    )


def _read_energy_dispersion(hdu):
    hdu
    return {
        "energy_bin_edges_GeV": bu.merge_low_high_edges(
            low=hdu.data["ENERG_LO"][0, :] * 1e3,
            high=hdu.data["ENERG_HI"][0, :] * 1e3,
        ),
        "Mu_bin_edges": bu.merge_low_high_edges(
            low=hdu.data["MIGRA_LO"][0, :],
            high=hdu.data["MIGRA_HI"][0, :],
        ),
        "theta_bin_edges_deg": bu.merge_low_high_edges(
            low=hdu.data["THETA_LO"][0, :],
            high=hdu.data["THETA_HI"][0, :],
        ),
        "dPdMu_vs_theta": hdu.data["MATRIX  "][0],
    }


def _read_effective_area(hdu):
    return {
        "energy_bin_edges_GeV": bu.merge_low_high_edges(
            low=hdu.data["ENERG_LO"][0, :] * 1e3,
            high=hdu.data["ENERG_HI"][0, :] * 1e3,
        ),
        "theta_bin_edges_deg": bu.merge_low_high_edges(
            low=hdu.data["THETA_LO"][0, :],
            high=hdu.data["THETA_HI"][0, :],
        ),
        "area_vs_theta_m2": hdu.data["EFFAREA"][0],
    }


def _read_background(hdu):
    per_MeV_to_per_GeV = 1e3
    return {
        "energy_bin_edges_GeV": bu.merge_low_high_edges(
            low=hdu.data["ENERG_LO"][0, :] * 1e3,
            high=hdu.data["ENERG_HI"][0, :] * 1e3,
        ),
        "detx_bin_edges_deg": bu.merge_low_high_edges(
            low=hdu.data["DETX_LO"][0, :],
            high=hdu.data["DETX_HI"][0, :],
        ),
        "dety_bin_edges_deg": bu.merge_low_high_edges(
            low=hdu.data["DETY_LO"][0, :],
            high=hdu.data["DETY_HI"][0, :],
        ),
        "background_vs_detx_vs_dety_per_s_per_sr_per_GeV": per_MeV_to_per_GeV
        * hdu.data["BKG"][0],
    }


def _read_point_spread_function(hdu):
    return {
        "energy_bin_edges_GeV": bu.merge_low_high_edges(
            low=hdu.data["ENERG_LO"][0, :] * 1e3,
            high=hdu.data["ENERG_HI"][0, :] * 1e3,
        ),
        "theta_bin_edges_deg": bu.merge_low_high_edges(
            low=hdu.data["THETA_LO"][0, :],
            high=hdu.data["THETA_HI"][0, :],
        ),
        "sigma_vs_theta_deg": hdu.data["SIGMA_1"][0],
    }


def read_instrument_response_function(path):
    irf = {}
    with astropy.io.fits.open(path) as f:
        irf["effective_area"] = _read_effective_area(hdu=f["EFFECTIVE AREA"])
        irf["energy_dispersion"] = _read_energy_dispersion(
            hdu=f["ENERGY DISPERSION"]
        )
        irf["background"] = _read_background(hdu=f["BACKGROUND"])
        irf["point_spread_function"] = _read_point_spread_function(
            hdu=f["POINT SPREAD FUNCTION"]
        )

    return irf


def average_instrument_response_over_field_of_view(irf, roi_opening_deg):
    assert roi_opening_deg > 0.0

    irf["effective_area"]["area_m2"] = average_field_of_view_radial(
        X=irf["effective_area"]["area_vs_theta_m2"],
        theta_bin_edges_deg=irf["effective_area"]["theta_bin_edges_deg"],
        roi_opening_deg=roi_opening_deg,
    )

    irf["energy_dispersion"]["dPdMu"] = average_field_of_view_radial(
        X=irf["energy_dispersion"]["dPdMu_vs_theta"],
        theta_bin_edges_deg=irf["energy_dispersion"]["theta_bin_edges_deg"],
        roi_opening_deg=roi_opening_deg,
    )

    irf["background"][
        "background_per_s_per_sr_per_GeV"
    ] = average_field_of_view_grid(
        X=irf["background"]["background_vs_detx_vs_dety_per_s_per_sr_per_GeV"],
        detx_bin_edges_deg=irf["background"]["detx_bin_edges_deg"],
        dety_bin_edges_deg=irf["background"]["dety_bin_edges_deg"],
        roi_opening_deg=roi_opening_deg,
    )

    irf["point_spread_function"]["sigma_deg"] = average_field_of_view_radial(
        X=irf["point_spread_function"]["sigma_vs_theta_deg"],
        theta_bin_edges_deg=irf["point_spread_function"][
            "theta_bin_edges_deg"
        ],
        roi_opening_deg=roi_opening_deg,
    )

    return irf


def integrate_dPdMu_to_get_probability_reco_given_true(
    dPdMu,
    dPdMu_energy_bin_edges,
    dPdMu_Mu_bin_edges,
    energy_bin_edges,
):
    """
    Convert a dispersion-matrux dPdMu to a conditional probability-matrix.

    dP/dMu is the probability-density-function and must fulfill:

    Int_0^\\inf \\frac{dP}{d\\mu} d\\mu = 1

    with \\mu being the ratio:

    \\mu = \\frac{E_{reco}}{E_{true}}.

    The conditional probability R to get E_{reco} given E_{true} is:

    R(i, j) = (\\Delta E_{true})^{-1} Int_\\Delta E_{true} R(i, E_{true}) dE_{true}

    with:

    R(i, E_true) = Int_\\mu(\\Delta E_{reco}) dP/d\\mu(E_true, \\mu) d\\mu

    Parameters
    ----------
    dPdMu : array(N, M)
        The dispersion-matrix.
    dPdMu_Mu_bin_edges : array(N)
        The true energy bin-edges used in dPdMu.
    dPdMu_energy_bin_edges : array(M)
        The margin's \\mu bin-edges used in dPdMu.
        This range sould be in awide range such as 0.2 < \\mu < 5.
    energy_bin_edges : array(J)
        The targeted energy-bin-edges of the conditional-probability-matrix
        to be computed.

    Returns
    -------
    conditional-probability-matrix : array(J, J)
        The conditional probability to measure E_{reco} given E_{true}.
        Using energy_bin_edges.
        ax0: true
        ax1: reco
    """
    assert bu.is_strictly_monotonic_increasing(dPdMu_Mu_bin_edges)
    assert bu.is_strictly_monotonic_increasing(dPdMu_energy_bin_edges)
    assert bu.is_strictly_monotonic_increasing(energy_bin_edges)

    assert np.all(dPdMu_Mu_bin_edges >= 0.0)
    assert np.all(dPdMu_energy_bin_edges >= 0.0)
    assert np.all(energy_bin_edges >= 0.0)

    assert np.all(dPdMu >= 0.0)

    dPdMu_Mu_bin_centers = bu.centers(dPdMu_Mu_bin_edges)
    Mu_bin_widths = bu.widths(dPdMu_Mu_bin_edges)
    dPdMu_energy_bin_centers = bu.centers(dPdMu_energy_bin_edges)
    dPdMu_energy_bin_widths = bu.widths(dPdMu_energy_bin_edges)
    energy_bin_centers = bu.centers(energy_bin_edges)

    N = len(dPdMu_Mu_bin_edges) - 1
    M = len(dPdMu_energy_bin_edges) - 1
    J = len(energy_bin_edges) - 1

    R = np.nan * np.ones(shape=(J, J))

    for r in range(J):
        for t in range(J):
            Etrue_center = energy_bin_centers[t]
            Ereco_start = energy_bin_edges[r]
            Ereco_stop = energy_bin_edges[r + 1]
            assert Ereco_stop > Ereco_start

            Mu_start = Ereco_start / Etrue_center
            Mu_stop = Ereco_stop / Etrue_center

            assert Mu_start < Mu_stop
            Mu_mask = np.logical_and(
                dPdMu_Mu_bin_centers >= Mu_start,
                dPdMu_Mu_bin_centers < Mu_stop,
            )

            # integrate over Mu
            # -----------------
            R_I_Etrue = np.zeros(dPdMu.shape[0])
            for u in range(N):
                R_I_Etrue += dPdMu[u, :] * Mu_mask[u] * Mu_bin_widths[u]

            assert R_I_Etrue.shape[0] == dPdMu.shape[0]

            Etrue_start = energy_bin_edges[t]
            Etrue_stop = energy_bin_edges[t + 1]
            assert Etrue_stop > Etrue_start
            Etrue_mask = np.logical_and(
                dPdMu_energy_bin_centers >= Etrue_start,
                dPdMu_energy_bin_centers < Etrue_stop,
            )
            delta_Etrue = Etrue_stop - Etrue_start

            # integrate over Etrue
            # --------------------
            R[t, r] = 0.0
            for et in range(M):
                R[t, r] += (
                    R_I_Etrue[et]
                    * Etrue_mask[et]
                    * dPdMu_energy_bin_widths[et]
                )
            R[t, r] /= delta_Etrue

    # normalize
    # ---------
    # ax0 -> true
    # ax1 -> reco
    probability_reco_given_true = np.nan * np.ones(shape=(J, J))

    for t in range(J):
        norm = np.sum(R[t, :])
        for r in range(J):
            probability_reco_given_true[t, r] = R[t, r] / norm

    # check
    for t in range(J):
        if np.sum(probability_reco_given_true[t, :]) > 0.0:
            assert 0.99 < np.sum(probability_reco_given_true[t, :]) < 1.01

    return probability_reco_given_true


def integrate_background_rate_in_onregion(
    background_per_s_per_sr_per_GeV,
    point_spread_function_sigma_deg,
    energy_bin_edges_GeV,
):
    psf_sigma_rad = np.deg2rad(point_spread_function_sigma_deg)
    energy_bin_widths_GeV = bu.widths(energy_bin_edges_GeV)

    background_rate_per_s = np.zeros(len(energy_bin_edges_GeV) - 1)
    for e in range(len(energy_bin_edges_GeV) - 1):
        psf_solid_angle_sr = np.pi * psf_sigma_rad[e] ** 2
        background_rate_per_s[e] = (
            background_per_s_per_sr_per_GeV[e]
            * energy_bin_widths_GeV[e]
            * psf_solid_angle_sr
        )
    return background_rate_per_s



import setuptools
import os

with open("README.rst", "r", encoding="utf-8") as f:
    long_description = f.read()


with open(os.path.join("flux_sensitivity", "version.py")) as f:
    txt = f.read()
    last_line = txt.splitlines()[-1]
    version_string = last_line.split()[-1]
    version = version_string.strip("\"'")


setuptools.setup(
    name="flux_sensitivity_sebastian-achim-mueller",
    version=version,
    description=(
        "Estimate the flux-sensitivity of a point-source either integral "
        "or differential w.r.t. energy."
    ),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/cherenkov-plenoscope/flux_sensitivity",
    author="Sebastian Achim Mueller",
    author_email="sebastian-achim.mueller@mpi-hd.mpg.de",
    packages=[
        "flux_sensitivity",
        "flux_sensitivity.io",
        "flux_sensitivity.integral",
        "flux_sensitivity.differential",
        "flux_sensitivity.differential.scenarios",
    ],
    package_data={
        "flux_sensitivity": [
            os.path.join("tests", "resources", "cta", "*"),
            os.path.join("tests", "resources", "portal", "*"),
        ]
    },
    install_requires=[
        "binning_utils_sebastian-achim-mueller>=0.0.5",
        "propagate_uncertainties_sebastian-achim-mueller>=0.2.2",
        "lima1983analysis_sebastian-achim-mueller>=0.0.4",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
    ],
)



import numpy as np


def init_matrix_B(probability_reco_given_true, containment=0.68):
    # ax0 -> true
    # ax1 -> reco
    num_bins = probability_reco_given_true.shape[0]
    M = probability_reco_given_true
    mask = np.zeros(shape=(num_bins, num_bins))

    # estimate containment regions:
    for etrue in range(num_bins):
        if np.sum(M[etrue, :]) > 0.0:
            assert 0.99 < np.sum(M[etrue, :]) < 1.01

            accumulated_containment = 0.0
            reco_best = np.argmax(M[etrue, :])

            accumulated_containment, weight = next_containment_and_weight(
                accumulated_containment=accumulated_containment,
                bin_containment=M[etrue, reco_best],
                target_containment=containment,
            )

            mask[etrue, reco_best] = weight
            start = reco_best - 1
            stop = reco_best + 1
            i = 0
            while accumulated_containment < containment:
                if start > 0:
                    accumulated_containment, w = next_containment_and_weight(
                        accumulated_containment=accumulated_containment,
                        bin_containment=M[etrue, start],
                        target_containment=containment,
                    )
                    mask[etrue, start] = w
                    start -= 1
                if accumulated_containment == containment:
                    break

                if stop + 1 < num_bins:
                    accumulated_containment, w = next_containment_and_weight(
                        accumulated_containment=accumulated_containment,
                        bin_containment=M[etrue, stop],
                        target_containment=containment,
                    )
                    mask[etrue, stop] = w
                    stop += 1
                if accumulated_containment == containment:
                    break

                if start == 0 and stop + 1 == num_bins:
                    break

                i += 1
                assert i < 2 * num_bins
    return mask


def next_containment_and_weight(
    accumulated_containment,
    bin_containment,
    target_containment,
):
    assert 0 <= accumulated_containment <= 1
    assert 0 <= bin_containment <= 1
    assert 0 < target_containment <= 1

    missing_containment = target_containment - accumulated_containment
    assert missing_containment > 0

    if bin_containment > 0:
        weight = np.min([missing_containment / bin_containment, 1])
    else:
        weight = 0

    if weight == 1:
        return accumulated_containment + bin_containment, 1
    else:
        return target_containment, weight
